options {language = Java;}





class LfP_Parser extends Parser;
options{
    k =3;
    exportVocab = LfP;
    defaultErrorHandler=true;
    buildAST = true;
}

tokens{

    ACCEPT = "accept";
    AND= "and";
    ARRAY = "array";
    ASYNCHRONOUS = "asynchronous";

    BAG;
    BEGIN = "begin";
    BINDER = "binder";
    BREAK = "break";

    CHANNEL = "channel";
    CIRCULAR = "circular";
    CLASS = "class";
    CONST = "const";

     DECLARE = "declare";

    ELSE = "else";
    ELSIF = "elsif";
    END = "end";
    ENUM = "enum";

    FALSE = "false";
    FIELD;
    FIFO = "fifo";
    FOR = "for";
    FUNCTION = "function";


    GOTO = "goto";

    IF = "if";
    IN = "in";
    INOUT = "inout" ;

    IS = "is";

    LABEL;

    NOT = "not";

    MEDIA = "media";

    NEW = "new";
    NULL_STATEMENT = "null";

    OF = "of";
    OPAQUE = "opaque";
    OR = "or";
    OUT = "out";

    PORT = "port";
    PROCEDURE = "procedure";

    RANGE = "range";
    RECORD = "record";
    RETURN = "return";
    ROOT;

    STATIC = "static";
    SYNCHRONOUS = "synchronous";
    SET = "set";

    THEN = "then";
    THIS = "this";
    TRIGGER = "trigger";
    TRUE = "true";
    TYPE = "type";

    VARIABLE;

    WHILE = "while";
    WITH = "with";


    UNARY_MINUS;



    EXPR_EXT_CALL; EXT_METHOD_NAME;

    NULL_EXPRESSION;



    TYPE_ATTRIBUTE;

    DECLARATION_IDENTIFIER;


}

lfp!: m:lfp_model { #lfp=#([ROOT,"ROOT"], #m); } ;

lfp_model:
        (type
        |
            (IDENTIFIER (COMA IDENTIFIER)* COLON IDENTIFIER (SEMI_COLON | AFFECT)) =>
            variable
        |
            (IDENTIFIER (COMA IDENTIFIER)* COLON IDENTIFIER WITH) =>
            static_instance
        |
            constant
        |
            media_declaration
        |
            class_declaration
        |
            binder_declaration
        |
            link_declaration
      )+
    ;

media_declaration: MEDIA^ IDENTIFIER (IS! (media_attribute)* (END! | body))? SEMI_COLON!
    ;


static_instance:
        IDENTIFIER^ (COMA IDENTIFIER)* COLON IDENTIFIER WITH
            OPEN_PAR (agregate_element)* CLOSE_PAR SEMI_COLON!
        ;
class_declaration: CLASS^ IDENTIFIER (IS! (attribute)* (END! | body))? SEMI_COLON!
    ;

attribute: type | variable | constant | methode | trigger
    ;
media_attribute:
        type | variable | constant | trigger
    ;


declaration: (type | variable | constant | trigger | methode)
    ;







body: BEGIN^ (statement)+ END!
    ;

type: TYPE^ IDENTIFIER IS! type_declaration SEMI_COLON!
    ;

type_declaration:
        ARRAY^ OPEN_PAR range (COMA range)* CLOSE_PAR OF IDENTIFIER
    |
        (CIRCULAR)? RANGE^ range OF! IDENTIFIER
    |
        (SET^ | BAG^) OF! IDENTIFIER
    |
        ENUM^ OPEN_PAR! IDENTIFIER (COMA! IDENTIFIER)* CLOSE_PAR!
    |
        PORT^ OPEN_PAR! IDENTIFIER (COMA! IDENTIFIER)* CLOSE_PAR!
    |
        OPAQUE^ ((external_method)+ END!)?
    |
        RECORD^ (field)+ END!
    ;


external_method:
        PROCEDURE^ IDENTIFIER OPEN_PAR! (parameters)? CLOSE_PAR! SEMI_COLON
    |
        FUNCTION^ IDENTIFIER OPEN_PAR! (parameters)? CLOSE_PAR! RETURN IDENTIFIER SEMI_COLON
    ;

field:
        IDENTIFIER (COMA! IDENTIFIER)* COLON^ IDENTIFIER SEMI_COLON!
        { #field = #([FIELD,"FIELD"], #field); }
    ;

variable: IDENTIFIER ( COMA! IDENTIFIER)* COLON^ IDENTIFIER (AFFECT expression)? SEMI_COLON!
        { #variable = #([VARIABLE,"VARIABLE"], #variable); }
    ;


constant!: CONST n:IDENTIFIER COLON! t:IDENTIFIER
        a:AFFECT e:expression SEMI_COLON
          { #constant = #(CONST, n, #(a, t, e)); }
    ;

trigger: TRIGGER^ IDENTIFIER OPEN_PAR (parameters)? CLOSE_PAR (RETURN^ IDENTIFIER)? (IS (variable | constant )? body)? SEMI_COLON!
    ;

methode: (SYNCHRONOUS FUNCTION) => function
        | (FUNCTION) => function
        | procedure
    ;

procedure:
        (ASYNCHRONOUS | SYNCHRONOUS)?
        PROCEDURE^ IDENTIFIER OPEN_PAR (parameters)? CLOSE_PAR
        (RIGHT_ARROW^ expression)?
        (IS^ (variable | constant)* body)? SEMI_COLON!
    ;

function: (SYNCHRONOUS)?
        FUNCTION^ IDENTIFIER OPEN_PAR! (parameters)? CLOSE_PAR!
        RETURN IDENTIFIER (RIGHT_ARROW^ expression)?
        (IS^ (variable | constant)* body)? SEMI_COLON!
    ;


parameters: parameter (SEMI_COLON! parameter)*
    ;

parameter: IDENTIFIER COLON^ (IN | INOUT | OUT)? IDENTIFIER
    ;

statement:
        if_then_else SEMI_COLON!
    |
        ((COLON IDENTIFIER)? FOR) => for_loop SEMI_COLON!
    |
        ((COLON IDENTIFIER)? WHILE) => while_loop SEMI_COLON!
    |
        break_statement SEMI_COLON!
    |
        (variable_occurence AFFECT) => affectation SEMI_COLON!
    |
        goto_statement SEMI_COLON!
    |
        ((COLON IDENTIFIER)? DECLARE) => declare_bloc SEMI_COLON!
    |
        NULL_STATEMENT SEMI_COLON!
    |
        set_label
    |
        (variable_occurence EXT_CALL) => external_call SEMI_COLON!
    |
        ( (discriminant)? OPEN_PAR expression (COMA expression)* CLOSE_PAR RIGHT_ARROW) => write_message SEMI_COLON!
    |
        ( (discriminant)? OPEN_PAR ) => read_message SEMI_COLON!
    |
        ( (discriminant)? IDENTIFIER ) => procedure_call SEMI_COLON!
    |
        (RETURN) => return_statement SEMI_COLON!
    |
        attente_activation SEMI_COLON!
    |
        accept_statement SEMI_COLON!
    ;

statements: (statement)+
    ;

if_then_else: IF^ expression THEN! statements (ELSIF^ expression THEN! statements)* (ELSE^ statements)? END!
    ;

for_loop: (COLON! IDENTIFIER)? FOR^ IDENTIFIER IN! range body
    ;

while_loop: (COLON! IDENTIFIER)? WHILE^ expression body
    ;

break_statement: BREAK^ (IDENTIFIER)?
    ;

goto_statement: GOTO^ IDENTIFIER
    ;

declare_bloc: (COLON! IDENTIFIER)? DECLARE^ (variable | constant)* body
    ;

attente_activation: IDENTIFIER LEFT_ARROW^ source_activation
  ;

source_activation: variable_occurence
  |
      OPEN_PAR CLOSE_PAR
  ;

accept_statement: ACCEPT^ attente_activation (COMA! attente_activation)* END!
  ;






binder_declaration: (STATIC)? BINDER^ f:IDENTIFIER DOT s:IDENTIFIER IS (FIFO | BAG) OPEN_PAR expression CLOSE_PAR readers writers END! SEMI_COLON!
    ;

readers: IDENTIFIER CAN_READ^ IDENTIFIER (COMA! IDENTIFIER)* SEMI_COLON!
    ;

writers: IDENTIFIER CAN_WRITE^ IDENTIFIER (COMA! IDENTIFIER)* SEMI_COLON!
    ;

link_declaration: MEDIA^ IDENTIFIER (linked_components | list_of_list) SEMI_COLON!
    ;

linked_components: OPEN_PAR^ IDENTIFIER (COMA IDENTIFIER)* CLOSE_PAR!
    ;

list_of_list: OPEN_PAR^ linked_components (COMA linked_components)* CLOSE_PAR!
    ;

range: expression RANGE_OP^ expression
    ;

affectation: variable_occurence AFFECT^ expression
    ;

set_label!:
        COLON! IDENTIFIER COLON!
        { #set_label=#([LABEL,"LABEL"],IDENTIFIER); }
    ;


read_message:
        (discriminant)? OPEN_PAR! expression (COMA expression)* CLOSE_PAR!
        LEFT_ARROW^ variable_occurence (WITH^ expression)?
    ;


write_message:
        (discriminant)? OPEN_PAR! expression (COMA expression)* CLOSE_PAR!
        RIGHT_ARROW^ variable_occurence
    ;

procedure_call:
        (discriminant)? (IDENTIFIER COLON)? IDENTIFIER
        OPEN_PAR (expression (COMA expression)*)? CLOSE_PAR
        (RIGHT_ARROW^ | ARROW^) variable_occurence
    ;

function_call:
        (discriminant)? (IDENTIFIER COLON)? IDENTIFIER OPEN_PAR (expression (COMA expression)*)? CLOSE_PAR
        EQUIV^ variable_occurence
    ;

external_call:

        variable_occurence EXT_CALL^ IDENTIFIER OPEN_PAR!
        (expression (COMA expression)*)? CLOSE_PAR!
    ;

discriminant:
        OPEN_BRACKET^ expression (COMA expression)* CLOSE_BRACKET!
    ;


return_statement:
        RETURN^ (expression)?
    ;

expression:
        simpleExpression
        ( (EQUAL^ | DIFF^ | LESS^ | LESS_EQ^ | GREATER^ | GREATER_EQ^) simpleExpression )*
    |
      IDENTIFIER QUOTE^ expression
    ;

simpleExpression:
        term ( (PLUS^ | MINUS^ | OR^) term )*
    ;

term:
    signedFactor ( (MULT^ | DIVIDE^ | AND^) signedFactor )*
    ;

signedFactor:

        (PLUS^ | MINUS^)? factor
    |
        NOT^ factor
    ;

factor:
        (variable_occurence EXT_CALL) => external_call
    |
        ((discriminant)? (IDENTIFIER COLON)? IDENTIFIER OPEN_PAR (expression (COMA expression)*)? CLOSE_PAR
        EQUIV) => function_call
    |
        (variable_occurence) => variable_occurence
    |
        INTEGER
    |
        OPEN_PAR^ expression CLOSE_PAR!
    |
        TRUE
    |
        FALSE
    |
        NULL_STATEMENT
    |
        dynamic_instanciation
    ;



variable_occurence:
        atomic_variable ( DOT atomic_variable)*
    ;

atomic_variable:
        IDENTIFIER^ (OPEN_PAR! (expression (COMA expression)*)? CLOSE_PAR!)?
    ;
dynamic_instanciation:
        NEW^ IDENTIFIER OPEN_PAR (agregate_element (COMA agregate_element)*)? CLOSE_PAR

    ;

agregate_element:
        IDENTIFIER IMPLIES^ expression
    ;


class LfP_Lexer extends Lexer;
options{
    k = 5;
    caseSensitiveLiterals = false;
    caseSensitive = false;
    charVocabulary='\u0000'..'\u00FE';
}


IDENTIFIER:
        ('a'..'z') ('a'..'z' | '_' | '0'..'9')*
    ;
INTEGER: ('0'..'9')+
    ;

NL: '\n'
        {

            $setType(Token.SKIP);



            newline();
        }
    ;

WS: (' ' | '\t' | '\r')+
        {

            $setType(Token.SKIP);



        }
    ;

COMMENT: "--"
        (~('\n'|'\r'))* ('\n'|'\r'('\n')?)

        {

            $setType(Token.SKIP);



            newline();
    }
  ;

AFFECT: ":=";
ARROW: "<->";
CAN_READ: "|->";
CAN_WRITE: "<-|";
CLOSE_BRACKET: "]";
CLOSE_PAR: ")";
COLON: ":";
COMA: ",";
DOT: ".";
EQUIV: "<=>";
IMPLIES: "=>";
LEFT_ARROW: "<-";
OPEN_BRACKET: "[";
OPEN_PAR:"(";
QUOTE: "'";
RANGE_OP: "..";
RIGHT_ARROW:"->";
SEMI_COLON: ";";

LESS: "<";
LESS_EQ: "<=";
EQUAL: "=";
DIFF: "/=";
GREATER_EQ: ">=";
GREATER: ">";

PLUS: "+";
MINUS: "-";

MULT: "*";
DIVIDE: "/";
EXT_CALL: "@";
